"""
StokeD
=======
Simulation and visualization of Stokesian motion
"""
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import quaternion

from . import analysis

from .solver import stokesian_dynamics, brownian_dynamics, interactions, trajectory
from .drag import drag, drag_sphere, drag_ellipsoid
from .inertia import inertia, inertia_sphere, inertia_ellipsoid
from .collisions import collisions_sphere, collisions_sphere_interface
from .electrostatics import (electrostatics, double_layer_sphere,
                             double_layer_sphere_interface)
from .electrodynamics import point_dipole_electrodynamics
from .van_der_waals import van_der_waals_sphere, van_der_waals_sphere_interface
from .gravity import gravity, gravity_sphere, gravity_ellipsoid
from .hydrodynamics import grand_mobility_matrix, interface
from .constraints import constrain_position, constrain_rotation
from .forces import pairwise_central_force, pairwise_force
from .common import lennard_jones

from .analysis import msd
from .utility import quaternion_to_angles
from .vis import trajectory_animation, circle_patches, ellipse_patches



from . import integrator
import numpy as np

class predictor_corrector_integrator(integrator):
    def __init__(self, grand_mobility_interval=10):
        super().__init__(grand_mobility_interval)

    def bd_step(self):
        r0 = np.copy(self.solver.position)
        o0 = np.copy(self.solver.orientation)

        F = self.solve_forces()
        Fr = self.random_force()
        F += Fr
        v1 = self.alpha_T*F

        self.solver.velocity = v1
        self.solver.position += self.dt*v1

        if self.solver.rotating:
            T = self.solve_torques()
            w1 = self.alpha_R*T
            w1_q = np.array([np.quaternion(*omega) for omega in w1])

            w0 = np.copy(self.solver.angular_velocity)
            o0 = np.copy(self.solver.orientation)
            self.solver.angular_velocity = w1
            self.solver.orientation = (1 + w1_q*self.dt/2)*self.solver.orientation

        self.solver.time += self.dt
        self.perform_constraints()
        self.pre_step()

        F = self.solve_forces()
        F += Fr
        v2 = self.alpha_T*F
        self.solver.velocity = (v1 + v2)/2
        self.solver.position = r0 + self.dt*self.solver.velocity

        if self.solver.rotating:
            T = self.solve_torques()
            w2 = self.alpha_R*T
            w2_q = np.array([np.quaternion(*omega) for omega in w1])
            w_q = (w1_q + w2_q)/2

            self.solver.angular_velocity = (w1 + w2)/2
            self.solver.orientation = (1 + w1_q*self.dt/2)*o0

        self.solver.time -= self.dt



import numpy as np
from stoked import interactions
from stoked.forces import pairwise_central_force

def collisions_sphere(radius, kn):
    def F(r):
        nonlocal radius

        if np.isscalar(radius):
            Nparticles = len(r)
            radius = np.full(Nparticles, radius, dtype=float)

        T1 = np.add.outer(radius, radius)

        overlap = T1 - r
        overlap[overlap<0] = 0
        return kn*overlap**1.5

    return pairwise_central_force(F)

class collisions_sphere_interface(interactions):
    """
    Collision between spheres and a planar surface using the force model F = kn*overlap^1.5
    """
    def __init__(self, radii, kn, zpos=0):
        """
        Arguments:
            radii   particle radii
            kn      force constant
            zpos    z-position of the plane (default: 0)
        """
        self.radii = radii
        self.kn = kn
        self.zpos = zpos

    def force(self):
        Nparticles = len(self.position)
        if np.isscalar(self.radii):
            rad = np.full(Nparticles, self.radii, dtype=float)
        else:
            rad = np.asarray(self.radii, dtype=float)

        F = np.zeros_like(self.position)

        dz = self.position[:,2] - self.zpos
        idx = np.abs(dz) < rad
        F[idx,2] = np.sign(dz[idx])*self.kn*np.sqrt((dz[idx] - rad[idx])**3)

        return F

    def torque(self):
        return np.zeros_like(self.position)



from abc import ABCMeta, abstractmethod
import numpy as np
from scipy.constants import k as kb
from stoked.hydrodynamics import particle_wall_self_mobility, grand_mobility_matrix
import quaternion

def fluctuation(friction, temperature, dt):
    """Given a friction matrix, return the fluctuation matrix"""
    if temperature == 0:
        return np.zeros_like(friction)
    else:
        rhs = 2*kb*temperature*friction/dt
        return np.linalg.cholesky(rhs)

def random_force_isotropic(alpha, temperature, dt):
    noise = np.random.normal(size=alpha.shape) 
    beta = np.sqrt(2*kb*temperature/(alpha*dt))
    return beta*noise

def random_force_anisotropic(alpha, temperature, dt):
    noise = np.random.normal(size=alpha.shape[:2]) 
    beta = np.zeros_like(alpha)
    for i in range(len(alpha)):
        beta[i] = np.linalg.inv(alpha[i]) @ fluctuation(alpha[i], temperature, dt)

    return np.einsum('Nij,Nj->Ni', beta, noise)

class integrator:
    __metaclass__ = ABCMeta

    def __init__(self, grand_mobility_interval=10):
        self.grand_mobility_interval = grand_mobility_interval
        self.total_steps = 0

    def initialize(self, solver):
        self.solver = solver
        self.dt = solver.dt
        self.hydrodynamics = solver.hydrodynamic_coupling
        self.isotropic = solver.drag.isotropic

    def solve_forces(self):
        F = self.solver._total_force(self.solver.time, self.solver.position, self.solver.orientation)
        return F

    def random_force(self):
        if self.isotropic:
            return random_force_isotropic(self.alpha_T, self.solver.temperature, self.dt)
        else:
            return random_force_anisotropic(self.alpha_T_rot, self.solver.temperature, self.dt)

    def random_torque(self):
        if self.isotropic:
            return random_force_isotropic(self.alpha_R, self.solver.temperature, self.dt)
        else:
            return random_force_anisotropic(self.alpha_R_rot, self.solver.temperature, self.dt)

    def solve_torques(self):
        T = self.solver._total_torque(self.solver.time, self.solver.position, self.solver.orientation)
        return T

    def perform_constraints(self):
        for constraint in self.solver.constraint:
            constraint(self.solver.position, self.solver.orientation)

    def pre_step(self):
        self.solver._update_interactions(self.solver.time, self.solver.position, self.solver.orientation)

        self._solve_mobility()
        if self.hydrodynamics and self.total_steps % self.grand_mobility_interval == 0:
            self._solve_grand_mobility()

        if not self.isotropic:
            self.solve_rotation_matrix()

    def post_step(self):
        if self.solver.rotating:
            self.solver.orientation = np.normalized(self.solver.orientation)

        self.solver.time += self.dt
        self.perform_constraints()
        self.total_steps += 1

    def solve_rotation_matrix(self):
        self.rot = quaternion.as_rotation_matrix(self.solver.orientation)
        self.alpha_T_rot = np.einsum('Nij,Nj,Nlj->Nil', self.rot, self.alpha_T, self.rot)
        self.alpha_R_rot = np.einsum('Nij,Nj,Nlj->Nil', self.rot, self.alpha_R, self.rot)

    def _solve_mobility(self):
        if self.solver.interface is not None:
            M_T = np.copy(self.solver.alpha_T)
            M_R = np.copy(self.solver.alpha_R)
            for i in range(self.solver.Nparticles):
                radius = self.solver.drag.radius[i] if self.solver.drag.radius.ndim else self.solver.drag.radius
                M_self = particle_wall_self_mobility(self.solver.position[i], self.solver.interface, self.solver.drag.viscosity, radius)
                M_T[i] -= M_self[0,0].diagonal()
                M_R[i] -= M_self[1,1].diagonal()
                self.alpha_T, self.alpha_R = M_T, M_R
        else:
            self.alpha_T, self.alpha_R = self.solver.alpha_T, self.solver.alpha_R

    def _solve_grand_mobility(self):
        drag_T = np.zeros([self.solver.Nparticles, 3, 3], dtype=float)
        drag_R = np.zeros([self.solver.Nparticles, 3, 3], dtype=float)
        np.einsum('Nii->Ni', drag_T)[...] = self.alpha_T
        np.einsum('Nii->Ni', drag_R)[...] = self.alpha_R

        self.grand_M = grand_mobility_matrix(self.solver.position, drag_T, drag_R, self.solver.drag.viscosity)
        self.grand_N = fluctuation(self.grand_M, self.solver.temperature, self.dt)

        # if self.drag.isotropic:
            # np.einsum('Nii->Ni', drag_T)[...] = self.alpha_T
            # np.einsum('Nii->Ni', drag_R)[...] = self.alpha_R
        # else:
            # rot = quaternion.as_rotation_matrix(orientation)
            # drag_T[...] = np.einsum('Nij,Nj,Nlj->Nil', rot, self.alpha_T, rot)
            # drag_R[...] = np.einsum('Nij,Nj,Nlj->Nil', rot, self.alpha_R, rot)

    @abstractmethod
    def bd_step(self):
        raise NotImplementedError('brownian dynamics step not implemented for this integrator')

    @abstractmethod
    def ld_step(self):
        raise NotImplementedError('langevin dynamics step not implemented for this integrator')

    @abstractmethod
    def hbd_step(self):
        raise NotImplementedError('hydrodynamic-brownian dynamics step not implemented for this integrator')

    @abstractmethod
    def hld_step(self):
        raise NotImplementedError('hydrodynamic-langevin dynamics step not implemented for this integrator')



from . import integrator
import numpy as np

class basic_integrator(integrator):
    def __init__(self, grand_mobility_interval=10):
        super().__init__(grand_mobility_interval)

    def bd_step(self):
        F = self.solve_forces()
        F += self.random_force()

        if self.isotropic:
            v1 = self.alpha_T*F
        else:
            v1 = np.einsum('Nij,Nj->Ni', self.alpha_T_rot, F)

        self.solver.velocity = v1
        self.solver.position += self.dt*v1

        if self.solver.rotating:
            T = self.solve_torques()
            T += self.random_torque()

            if self.isotropic:
                w1 = self.alpha_R*T
            else:
                w1 = np.einsum('Nij,Nj->Ni', self.alpha_R_rot, T)

            self.solver.angular_velocity = w1
            w1_q = np.array([np.quaternion(*omega) for omega in w1])
            self.solver.orientation = (1 + w1_q*self.dt/2)*self.solver.orientation

    def ld_step(self):
        if self.solver.rotating:
            raise NotImplementedError('ld step with rotation')

        if not self.isotropic:
            raise NotImplementedError('ld step with anisotropic particles')

        F = self.solve_forces()
        Fr = self.random_force()
        F += Fr

        mass = self.solver.mass
        v0 = self.solver.velocity
        v1 = self.alpha_T*F + (v0 - self.alpha_T*F)*np.exp(-self.dt/mass/self.alpha_T)

        dr = self.alpha_T*F*self.dt - self.alpha_T*mass*(self.alpha_T*F - v0)*(1 - np.exp(-self.dt/mass/self.alpha_T))

        self.solver.velocity = v1
        self.solver.position += dr

    def hbd_step(self):
        if not self.isotropic:
            raise NotImplementedError('hbd step with anisotropic particles')

        F = self.solve_forces()
        T = self.solve_torques()

        grand_F = np.hstack([F.flatten(), T.flatten()]) 
        grand_noise = np.random.normal(size=grand_F.shape)

        Np = self.solver.Nparticles
        grand_velocity = self.grand_M @ grand_F + self.grand_N @ grand_noise
        v1 = grand_velocity[:3*Np].reshape([Np,3])
        w1 = grand_velocity[3*Np:].reshape([Np,3])

        self.solver.velocity = v1
        self.solver.position += self.dt*v1

        self.solver.angular_velocity = w1
        w1_q = np.array([np.quaternion(*omega) for omega in w1])
        self.solver.orientation = (1 + w1_q*self.dt/2)*self.solver.orientation



"""
Functions for post-processing analysis of trajectory data
"""

import numpy as np
from scipy.integrate import cumtrapz
from tqdm import tqdm

def autocorrelation_fft(x):
    """
    compute the autocorrelation of x using the FFT
    """
    N = len(x)
    F = np.fft.fft(x, n=2*N)
    PSD = F*F.conjugate()
    res = np.fft.ifft(PSD)
    res = res[:N].real
    n = N*np.ones(N) - np.arange(N)

    return res/n

def msd_single(trajectory, fft=True):
    """
    Compute the mean-squared displacement (MSD) of a single trajectory

    Arguments:
        trajectory    shape [T,D] for a single trajectory of T steps in D dimensions
        fft           use the FFT to compute the MSD (default: True)
    """
    if fft:
        N, dim = trajectory.shape

        D = np.sum(np.square(trajectory), axis=1)
        D = np.append(D, 0)
        S2 = sum([autocorrelation_fft(trajectory[:,i]) for i in range(dim)])

        Q = 2*np.sum(D)
        S1 = np.zeros(N, dtype=float)
        for m in range(N):
            Q = Q - D[m-1] - D[N-m]
            S1[m] = Q/(N-m)

        return S1 - 2*S2
    else:
        Nsteps = len(trajectory)
        lags = np.arange(Nsteps)
        msd_vals = np.zeros(Nsteps, dtype=float)    

        for i, lag in enumerate(lags):
            dr = trajectory[:-lag if lag else None] - trajectory[lag:]
            dr2 = np.square(dr).sum(axis=1)
            msd_vals[i] = dr2.mean()

        return msd_vals

def msd(trajectory, fft=True):
    """
    Compute the mean-squared displacement (MSD) of a trajectory

    Arguments:
        trajectory    shape [T,D] or [T,N,D] for a trajectory of T steps, N particles in D dimensions
        fft           use the FFT to compute the MSD (default: True)
    """
    if trajectory.ndim == 2:
        return msd_single(trajectory, fft)

    elif trajectory.ndim == 3:
        T, N, D = trajectory.shape
        ret = np.zeros([T,N], dtype=float)
        for i in range(N):
            ret[:,i] = msd_single(trajectory[:,i], fft)
        return ret

    else:
        raise ValueError('trajectory does not have the appropriate shape (ndim needs to be 2 or 3)')

def angle_of_cluster(trajectory, wrap=False, initial=0, N_cutoff=1):
    """Determine the angle in time of a cluster of particles that is approximately a rigid structure
    
    Arguments:
        trajectory[T,N,D]        trajectory of N particles of length T (dimension D > 1)
        wrap       (bool) wrap the angles between -π and +π (default: False)
        initial    initial angle of the cluster (default: 0)
        N_cutoff   number of particles closest to the center of mass to ignore (default: 1)
    """
    Tsteps = trajectory.shape[0]
    
    ### obtain the trajectory in 2 dimensions centered around the center of mass
    com = np.average(trajectory[...,:2], axis=1)
    traj_2d = trajectory[...,:2] - com[:,np.newaxis]

    ### calculate the distance to center of mass for each particle at each time-step
    dist_to_com = np.linalg.norm(traj_2d, axis=-1)

    ### obtain 'trackers': particle indices that are furthert from the center of mass at time t
    trackers = np.argsort(dist_to_com, axis=-1)[:,N_cutoff:]

    ### calculate differential angle change of trackers at time t
    pos = np.array([traj_2d[i,trackers[i]] for i in range(Tsteps-1)])
    pos_next = np.array([traj_2d[i+1,trackers[i]] for i in range(Tsteps-1)])

    # phi = np.arctan2(pos[...,1], pos[...,0])
    # phi_next = np.arctan2(pos_next[...,1], pos_next[...,0])
    # dphi = phi_next - phi
    # dphi = np.sign(dphi) * (np.abs(dphi) % np.pi)

    dr = pos_next - pos
    dphi = np.cross(pos, dr, axis=-1)/np.linalg.norm(pos, axis=-1)**2

    ### the angle of the cluster is then an integrated average of dphi
    dphi_avg = np.average(dphi, axis=1)    # average of trackers
    phi_cluster = cumtrapz(dphi_avg, initial=initial)    # angle integral of differential angles
    phi_cluster = np.insert(phi_cluster, 0, 0, axis=0)

    if not wrap:
        phi_cluster = np.unwrap(phi_cluster)

    return phi_cluster

def transform_to_lattice(trajectory, lattice):
    """
    Returns a new trajectory that is translated and rotated to be best aligned with a given fixed lattice
    
    Arguments:
        trajectory[T,N,2]      trajecrory data with T steps, N particles in 2 dimensions
        lattice[N,2]           lattice to align to
    """
    Nsteps, Nparticles, _ = trajectory.shape

    translation = np.average(trajectory[...,:2], axis=1)
    P = trajectory - translation[:,np.newaxis]
    Q = lattice[:,:2]

    H = np.einsum('TNi,Nj->Tij', P, Q)
    U, S, V = np.linalg.svd(H)
    R = np.einsum('Tij,Tjk->Tik', V, U)

    d = np.linalg.det(R)

    new_trajectory = np.einsum('Tij,TNj->TNi', R, P)

def transform_to_lattice_2d(trajectory, lattice):
    """
    Returns a new trajectory that is translated and rotated to be best aligned with a given fixed lattice in 2 dimensions
    
    Arguments:
        trajectory[T,N,2]      trajecrory data with T steps, N particles in 2 dimensions
        lattice[N,2]           lattice to align to
    """
    Nsteps, Nparticles, _ = trajectory.shape

    z1 = trajectory[...,0] + 1j*trajectory[...,1]
    z2 = lattice[...,0] + 1j*lattice[...,1]

    az1 = np.average(z1, axis=1)
    az2 = np.average(z2, axis=0)

    z = az2 - az1
    translation = np.array((z.real, z.imag)).T

    z1 -= az1[:,np.newaxis]
    z2 -= az2
    a = np.linalg.norm(z1, axis=1)**2

    err = np.linalg.norm(z2)**2
    zz1 = np.einsum('TN,N->T', z1, z2)
    zz_1 = np.einsum('TN,N->T', z1, np.conj(z2))

    b1 = np.abs(zz1)
    b_1 = np.abs(zz_1)
    err1 = -b1**2/a
    err_1 = -b_1**2/a

    idx = err1 <= err_1

    mirr = np.empty(Nsteps, dtype=float)
    sc = np.empty(Nsteps, dtype=complex)
    theta = np.empty(Nsteps, dtype=float)
    error = err*np.ones(Nsteps, dtype=float)
    rot = err*np.ones(Nsteps, dtype=complex)

    mirr[idx] = 1
    error[idx] += err1[idx]
    sc[idx] = b1[idx]/a[idx]
    theta[idx] = -np.angle(zz1[idx])
    rot[idx] = zz1[idx]

    mirr[~idx] = -1
    error[~idx] += err_1[~idx]
    sc[~idx] = b_1[~idx]/a[~idx]
    theta[~idx] = -np.angle(zz_1[~idx])
    rot[~idx] = zz_1[~idx]

    z1 = np.einsum('TN,T->TN', z1, np.conj(rot)/np.abs(rot))
    new_trajectory = np.empty_like(trajectory)
    new_trajectory[...,0] = z1.real
    new_trajectory[...,1] = z1.imag
    error = np.sqrt(error)

    return new_trajectory

def linear_stability_analysis(bd):
    """
    Perform a linear stability analysis for a given brownian dynamics simulation.
    Returns (eigenvalues, eigenvectors)
    """
    # bd.run(2000)

    F = bd._total_force(bd.time, bd.position, bd.orientation)
    com = np.average(bd.position, axis=0)
    T = np.cross(bd.position - com, F, axis=1)[:,2]

    dx = 1e-11
    force_matrix = np.zeros([bd.Nparticles, 2, bd.Nparticles, 2], dtype=float)
    for i in tqdm(range(bd.Nparticles), desc='Linear stability analysis K-matrix'):
        for j in [0,1]:
            bd.position[i,j] += dx
            bd._update_interactions(bd.time, bd.position, bd.orientation)
            F = bd._total_force(bd.time, bd.position, bd.orientation)
            bd.position[i,j] -= dx

            for k in range(len(F)):
                r = np.linalg.norm(bd.position[k] - com)
                tangent = np.array([-bd.position[k][1] - com[1], bd.position[k][0] - com[0], 0])
                if r > dx:
                    F[k] -= T[k]/r**2*tangent

            force_matrix[...,i,j] = F[:,:2]/dx

    w, v = np.linalg.eig(force_matrix.reshape([2*force_matrix.shape[0], -1]))
    v = v.reshape([bd.Nparticles, 2, 2*bd.Nparticles])
    v = np.moveaxis(v, -1, 0)

    return w, v

def psi_n(trajectory, n, rmax, single=None):
    """
    Compute psi_n, a measure of the discrete rotational symmetry

    Arguments:
        trajectory      particle trajectory
        n               degree of rotational symmetry
        rmax            maximum distance that defines nearest neighbors
        single          index of a single particle to compute psi_n for (default: average over all)
    """
    Nsteps = len(trajectory)
    Nparticles = trajectory.shape[1]

    psi = np.zeros(Nsteps, dtype=complex)

    if single is None:
        for i in range(Nsteps):
            for j in range(Nparticles):
                r_ij = trajectory[i,j,:2] - np.delete(trajectory[i,:,:2], j, axis=0)
                idx = np.linalg.norm(r_ij, axis=1) < rmax
                theta_ij = np.arctan2(r_ij[idx,1], r_ij[idx,0])

                if len(theta_ij) != 0:
                    psi[i] += np.sum(np.exp(1j*n*theta_ij))/(Nparticles*len(theta_ij))
    else:
        for i in range(Nsteps):
            r_ij = trajectory[i,single,:2] - np.delete(trajectory[i,:,:2], single, axis=0)
            idx = np.linalg.norm(r_ij, axis=1) < rmax
            theta_ij = np.arctan2(r_ij[idx,1], r_ij[idx,0])

            if len(theta_ij) != 0:
                psi[i] = np.sum(np.exp(1j*n*theta_ij))/(len(theta_ij))

    return psi

def radial_distrubtion(pos, bins=100, range=None, weighted=False, dim=None):
    """
    Compute a radial distribution function. Returns (counts, bin_edges)

    Arguments:
        pos[T,N,ndim]     particle positions
        bins              number of bins in histrogram
        range             lower and upper range of the bins
        weighted          if True, weight the RDF (default: False)
        dim               if weighted, the dimensionality to use (default: infer from pos)
        **kwargs          additional kwargs to np.histogram
    """
    Ntime, Nparticles, Ndim = pos.shape
    if dim is None:
        dim = Ndim

    dr = pos[:,np.newaxis] - pos[:,:,np.newaxis]
    dr_norm = np.linalg.norm(dr, axis=-1)
    dr_norm = dr_norm.flatten()
    idx = dr_norm == 0
    dr_norm = dr_norm[~idx]

    weights = 1/dr_norm**(dim-1) if weighted else None
    return np.histogram(dr_norm, bins=bins, weights=weights, density=True, range=range)

def radial_distrubtion_pair(p1, p2, bins=100, range=None, weighted=False, dim=None):
    """
    Compute a pair radial distribution function. Returns (counts, bin_edges)

    Arguments:
        p1[T,N1,ndim]     particle group 1 positions
        p1[T,N2,ndim]     particle group 2 positions
        bins              number of bins in histrogram
        range             lower and upper range of the bins
        weighted          if True, weight the RDF (default: False)
        dim               if weighted, the dimensionality to use (default: infer from pos)
        **kwargs          additional kwargs to np.histogram
    """
    Ntime, Nparticles, Ndim = p1.shape
    if dim is None:
        dim = Ndim

    dr = p1[:,np.newaxis] - p2[:,:,np.newaxis]
    dr_norm = np.linalg.norm(dr, axis=-1)
    dr_norm = dr_norm.flatten()

    weights = 1/dr_norm**(dim-1) if weighted else None
    return np.histogram(dr_norm, bins=bins, weights=weights, density=True, range=range)



"""
Electrodynamic interactions in the point-dipole approximation
"""
import numpy as np
from scipy import constants
import quaternion
from stoked import interactions


def greens(R, k, eps_b=1):
    """Return the Green's matrix"""
    r = np.linalg.norm(R)
    A = np.exp(1j*k*r)/(4*np.pi*constants.epsilon_0*eps_b*r**3)
    B = (3 - 3j*k*r - k**2*r**2)/r**2 * np.outer(R,R)
    C = (k**2*r**2 + 1j*k*r - 1) * np.identity(3)
    return A*(B + C)

def greens_any_dim(R, k, eps_b=1, delta=0):
    """Return the Green's matrix"""
    # R shape: [...,3]
    r = np.linalg.norm(R, axis=-1) + delta # shape [...]
    A = np.exp(1j*k*r)/(4*np.pi*constants.epsilon_0*eps_b*r**3)  # shape [...]
    B = np.einsum('...,...i,...j->...ij', (3 - 3j*k*r - k**2*r**2)/r**2, R, R)  # shape [...,3,3]
    C = np.einsum('...,ij->...ij', k**2*r**2 + 1j*k*r - 1, np.identity(3))  # shape [...,3,3]
    return np.einsum('...,...ij->...ij', A, B + C)
    #return np.einsum(A, [Ellipsis], B+C, [Ellipsis, 0, 1])

class point_dipole_electrodynamics(interactions):
    """Perform point dipole electrodynamic interactions"""
    def __init__(self, alpha_0, source, wavelength, eps_b=1, interactions=True, use_dynamic_polarizability=True):
        """Arguments:
                alpha_0[N,3]    static polarizability vectors of N particles (alternative shape: [3], CONST)
                source          incident source object that defines E and H functions
                wavelength      wavelength of illumination (in meters)
                eps_b           background permitivitty (default: 1)
                interactions    include particle interactions (default: True)
                use_dynamic_polarizability       modify the input alpha to be frequency corrected
        """

        self.Nparticles = len(alpha_0)
        self.alpha_0 = np.zeros([self.Nparticles,3,3], dtype=complex)
        np.einsum('Nii->Ni', self.alpha_0)[...] = alpha_0

        self.source = source
        self.wavelength = wavelength
        self.eps_b = eps_b
        self.k = 2*np.pi*self.eps_b**0.5/self.wavelength
        self.interactions = interactions

        self.sol = None
        self.G = np.zeros(shape = (self.Nparticles, 3, self.Nparticles, 3), dtype=complex)

        if use_dynamic_polarizability:
            # self.alpha = self.alpha_0/(1 - 2/3*1j*self.k**3*self.alpha_0)
            self.alpha = self.alpha_0/(1 - 1j*self.k**3*self.alpha_0/(6*np.pi*eps_b*constants.epsilon_0))
            self.alpha = self.alpha_0/(1 - 1j*self.k**3*self.alpha_0/(6*np.pi*eps_b*constants.epsilon_0)
                    - self.k**2*self.alpha_0/(6*np.pi*eps_b*constants.epsilon_0*(75e-9)))
        else:
            self.alpha = self.alpha_0

        self.alpha_0_t = np.copy(self.alpha_0)
        self.alpha_t = np.copy(self.alpha)

    def solve(self):
        """Solve for the electric fields given the current state of particles"""

        Einc = self.source.E_field(*self.position.T, self.k).T

        rot = quaternion.as_rotation_matrix(self.orientation)
        self.alpha_t[...] = np.einsum('Nij,Njk,Nlk->Nil', rot, self.alpha, rot)
        self.alpha_0_t[...] = np.einsum('Nij,Njk,Nlk->Nil', rot, self.alpha_0, rot)

        if not self.interactions:
            self.sol = Einc
        else:
            identity = np.zeros(shape = (self.Nparticles, 3, self.Nparticles, 3), dtype=complex)
            np.einsum('ixix->xi', identity)[...] = 1

            for i in range(self.Nparticles):
                for j in range(self.Nparticles):
                    if i == j: continue
                    pi = self.position[i]
                    pj = self.position[j]
                    dji = pi -  pj
                    self.G[i,:,j,:] = greens(dji, self.k, self.eps_b)
                    # G[j,:,i,:] = G[i,:,j,:]

            A = identity - np.einsum('ixjy,jyz->ixjz', self.G, self.alpha_t)
            self.sol = np.linalg.tensorsolve(A, Einc)

    def Efield(self, r, inc=True, delta=1e-13):
        """Compute the electic field at positions r[...,3] due to all particles

           Arguments:
                 r         positions (shape: [...,3])
                 inc       include the incident field (default: True)
                 delta     pad the distance between point and scatterers by delta (to avoid singularities)
        """

        if inc:
            E = self.source.E_field(*r.T, self.k).T
        else:
            E = np.zeros_like(r, dtype=np.complex)

        for i in range(self.Nparticles):
            R = r - self.particles.position[i]
            G = greens_any_dim(R, self.k, self.eps_b, delta)
            E += np.einsum('...xy,yz,z', G, self.alpha_t[i], self.sol[i]) 
        return E

    def update(self):
        self.solve()

    def force(self):
        """Return the force and torque on all particles, (F[N,3],T[N,3])"""

        def Efield_without(r, i):
            """Compute Efield without particle i"""
            E = self.source.E_field(*r, self.k)
            for j in range(self.Nparticles):
                if i == j: continue
                R = r - self.position[j]
                G = greens(R, self.k, self.eps_b)
                E += np.einsum('xy,yz,z', G, self.alpha_t[j], self.sol[j]) 
            return E

        F = np.zeros([self.Nparticles, 3])

        eps = 1e-12
        for i in range(self.Nparticles):
            Edx = Efield_without(self.position[i] + eps*np.array([1,0,0]), i)
            Edy = Efield_without(self.position[i] + eps*np.array([0,1,0]), i)
            Edz = Efield_without(self.position[i] + eps*np.array([0,0,1]), i)
            dE = (np.array([Edx,Edy,Edz]) - self.sol[i])/eps

            F[i] = 0.5*np.real(
                np.einsum('xy,y,zx', np.conj(self.alpha_t[i]), np.conj(self.sol[i]), dE))

        return F

    def torque(self):
        """Return the force and torque on all particles, (F[N,3],T[N,3])"""
        T = np.zeros([self.Nparticles, 3])

        for i in range(self.Nparticles):
            p = np.dot(self.alpha_t[i],self.sol[i])
            a0_inverse = np.linalg.inv(self.alpha_0_t[i])
            T[i] = 0.5*np.real(
                np.cross(np.conj(p), np.dot(a0_inverse, p)))

        return T



import numpy as np
from stoked import interactions
from stoked.forces import pairwise_central_force

def van_der_waals_sphere(radius, hamaker):
    radius = np.asarray(radius, dtype=float)

    def F(r):
        nonlocal radius

        Nparticles = len(r)
        if not np.ndim(radius):
            radius = np.full(Nparticles, radius, dtype=float)

        factor = 2*self.hamaker/3
        T1 = np.outer(radius, radius)
        T2 = np.add.outer(radius, radius)**2
        T3 = np.substract.outer(radius, radius)**2
        Q = r*T1*(1/(r**2 - T2**2) - 1 /(r**2 - T3**2))**2

        return Q

    return F

class van_der_waals_sphere_interface(interactions):
    def __init__(self, radius, hamaker, zpos=0):
        self.radius = np.asarray(radius, dtype=float)
        self.hamaker = hamaker
        self.zpos = zpos

    def force(self):
        Nparticles = len(self.position)
        if not np.ndim(self.radius):
            radius = np.full(Nparticles, self.radius, dtype=float)
        else:
            radius = self.radius

        factor = self.hamaker/6
        dz = self.position[:,2] - self.zpos

        Q = radius*(1/(dz - radius) - 1/(dz + radius))**2

        F = np.zeros_like(self.position)
        F[:,2] = -np.sign(dz)*factor*Q
        return F

    def torque(self):
        return np.zeros_like(self.position)



"""
Utility functions
"""
import numpy as np
import quaternion
from scipy.integrate import cumtrapz

def quaternion_to_angles(quat, reference=None):
    """
    Convert a quaternion array to an angle representation

    Arguments:
        quat        [T,...] quaternion trajectory of T time-steps
        reference   reference frame (as quaternion) around which to compute angles (default: z-axis)
    """
    if reference is not None:
        quat = np.invert(reference)*quat

    ### calculates differntial angle at each time-step, and cumtrapz to obtain angle
    quat_s = np.roll(quat, 1, axis=0)
    Q = quat*np.invert(quat_s)
    axis_angle = quaternion.as_rotation_vector(Q)
    d_angle = axis_angle[...,2]
    d_angle[0] = 0    # first entry is unphysical, so set to 0

    ### obtain the initial angles; multiply phi by 2 if theta = 0 for proper conversion
    theta, phi = np.moveaxis(quaternion.as_spherical_coords(quat[0]), -1, 0)
    idx = (theta == 0)
    phi[idx] *= 2

    angle = phi + cumtrapz(d_angle, axis=0, initial=0)
    return angle



import numpy as np
from scipy import constants
from stoked import interactions
from stoked.forces import pairwise_central_force

def electrostatics(charges):
    """
    Free-space point electrostatic interactions
    """
    ke = 1/(4*np.pi*constants.epsilon_0)

    def F(r):
        if np.isscalar(charges):
            Nparticles = len(r)
            Q = np.full(Nparticles, charges, dtype=float)
        else:
            Q = np.asarray(charges, dtype=float)

        return ke*np.outer(Q, Q)/r**2

    return pairwise_central_force(F)

def double_layer_sphere(radius, potential, temperature=300, debye=27.6e-9, eps_m=80.4, zp=1):
    """
    Electrostatic interactions in a fluid medium for spheres
    """
    radius = np.asarray(radius, dtype=float)
    potential = np.asarray(potential, dtype=float)

    def F(r):
        nonlocal radius, potential
        Nparticles = len(r)
        if not np.ndim(radius):
            radius = np.full(Nparticles, radius, dtype=float)

        if not np.ndim(potential):
            potential = np.full(Nparticles, potential, dtype=float)

        factor = 16*np.pi*constants.epsilon_0*eps_m/debye \
                 *(constants.k*temperature/(zp*constants.elementary_charge))**2
    
        T1 = np.add.outer(radius, radius)
        if temperature == 0:
            T2 = np.ones_like(potential)
        else:
            T2 = np.tanh(zp*constants.elementary_charge*potential/(4*constants.k*temperature))
        T2 = np.outer(T2, T2)

        Q = factor*T1*T2*np.exp(-(r - T1)/debye)
        return Q

    return pairwise_central_force(F)

class double_layer_sphere_interface(interactions):
    """
    Electrostatic interactions in a fluid medium near an interface
    """
    def __init__(self, radius, potential, potential_interface, debye=27.6e-9, eps_m=80.4, zp=1, zpos=0):
        self.radius = np.asarray(radius, dtype=float)
        self.potential = np.asarray(potential, dtype=float)
        self.potential_interface = potential_interface
        self.debye = debye
        self.eps_m = eps_m
        self.zp = zp
        self.zpos = zpos

    def force(self):
        Nparticles = len(self.position)
        if not np.ndim(self.radius):
            radius = np.full(Nparticles, self.radius, dtype=float)
        else:
            radius = self.radius
        if not np.ndim(self.potential):
            potential = np.full(Nparticles, self.potential, dtype=float)
        else:
            potential = self.potential

        factor = 16*np.pi*constants.epsilon_0*self.eps_m/self.debye \
                 *(constants.k*self.temperature/(self.zp*constants.elementary_charge))**2

        dz = self.position[:,2] - self.zpos

        T1 = np.tanh(self.zp*constants.elementary_charge*self.potential_interface/(4*constants.k*self.temperature))
        if self.temperature == 0:
            T2 = np.ones_like(potential)
        else:
            T2 = np.tanh(self.zp*constants.elementary_charge*potential/(4*constants.k*self.temperature))
        Q = np.sign(dz)*factor*radius*T1*T2*np.exp(-(np.abs(dz) - radius)/self.debye)

        F = np.zeros_like(self.position)
        F[:,2] = Q
        return F

    def torque(self):
        return np.zeros_like(self.position)



import numpy as np
from abc import ABCMeta, abstractmethod

class inertia:
    """
    Abstract base class for drag coeffecients
    """
    __metaclass__ = ABCMeta

    def __init__(self, density, isotropic=False):
        """
        Arguments:
            density        mass density of the object
            isotropic      (bool) True if the moment of inertia is isotropic
        """
        self.density = density
        self.isotropic = isotropic

    @abstractmethod
    def _mass(self):
        raise NotImplementedError('mass has not been implemented for this type of particle')

    @abstractmethod
    def _moment(self):
        raise NotImplementedError('moment has not been implemented for this type of particle')

    @property
    def mass(self):
        """particle mass"""
        return self._mass()

    @property
    def moment(self):
        """particle moment of inertia"""
        return self._moment()


class inertia_sphere(inertia):
    """
    Inertia values for a sphere
    """

    def __init__(self, radius, density):
        """
        Arguments:
            radii[N,3]     ellipsoid radii
            density        mass density of the object
        """
        super().__init__(density, isotropic=True)
        self.radius = np.asarray(radius, dtype=float)

    def _mass(self):
        return 4/3*np.pi*self.radius**3*self.density

    def _moment(self):
        M = self.mass
        return 2/5*M*self.radius**2

class inertia_ellipsoid(inertia):
    """
    Inertia values for an ellipsoid
    """

    def __init__(self, radii, density):
        """
        Arguments:
            radii[N,3]     ellipsoid radii
            density        mass density of the object
        """
        super().__init__(density, isotropic=True)
        self.radii = np.atleast_2d(np.asarray(radii, dtype=float))

    def _mass(self):
        V = 4/3*np.pi*np.product(self.radii, axis=1)
        return V*self.density

    def _moment(self):
        M = self.mass
        Ix = 1/5*M*(self.radii[:,1]**2 + self.radii[:,2]**2)
        Iy = 1/5*M*(self.radii[:,0]**2 + self.radii[:,2]**2)
        Iz = 1/5*M*(self.radii[:,0]**2 + self.radii[:,1]**2)

        return np.array([Ix, Iy, Iz]).T



import numpy as np
import quaternion

def constrain_position(x=None, y=None, z=None, idx=None):
    if idx is None:
        idx = np.s_[:]

    idy = []
    vals = []
    if x is not None:
        idy.append(0)
        vals.append(x)

    if y is not None:
        idy.append(1)
        vals.append(y)

    if z is not None:
        idy.append(2)
        vals.append(z)

    index = np.s_[idx,idy]
    vals = np.asarray(vals)

    def contraint(position, orientation):
        position[index] = vals

    return contraint

def constrain_rotation(x=False, y=False, z=False):
    idy = []
    if x:
        idy.append(0)

    if y:
        idy.append(1)

    if z:
        idy.append(2)


    def contraint(position, orientation):
        for i in range(len(orientation)):
            vec = orientation[i].vec
            vec[idy] = 0
            q = quaternion.quaternion(orientation[i].real, *vec)
            orientation[i] = q.normalized()

    return contraint



def external_force_potential(Ufunc):
    def force(time, position, orientation):
        pass

    return force

def external_torque_potential(Ufunc):
    def torque(time, position, orientation):
        pass

    return torque



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

NAME = 'stoked'
DESCRIPTION = "Simulation and visualization of Stokesian dynamics for N interacting particles"
URL = ''
EMAIL = 'japarker@uchicago.edu'
AUTHOR = 'John Parker'
KEYWORDS = 'stokesian dynamics brownian'
# REQUIRES_PYTHON = '>=3.6.0'
VERSION = '0.3.0'
LICENSE = 'MIT'

REQUIRED = [
    'numpy', 
    'scipy',
    'matplotlib',
    'tqdm',
    'numpy_quaternion',
]

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    keywords=KEYWORDS,
    url=URL,
    packages=find_packages(),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    install_requires=REQUIRED,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Physics',
        'Intended Audience :: Science/Research',
    ],
    zip_safe=False,
)



from stoked.forces import pairwise_central_force

def lennard_jones(rmin, eps):
    """Lennard-Jones pairwise interactions

    Arguments:
        rmin       distance where potential is a minimum 
        eps        depth of the potential well
    """

    def F(r):
        return 12*eps*(rmin**12/r**13 - rmin**6/r**7)

    return pairwise_central_force(F)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection

class colored_line:
    def __init__(self, x, y, c, **kwargs):
        """Small wrapper for a LineCollection with easier to use set_data functions
               x[N]                   xdata
               y[N]                   ydata
               c[N], c[N,3], c[N,4]   color data
               **kwargs               kwargs valid for matplotlib.LineCollection
        """
        x = np.asarray(x)
        y = np.asarray(y)
        c = np.asarray(c)

        xy = np.vstack([x,y]).T
        xy = xy.reshape(-1, 1, 2)
        segments = np.hstack([xy[:-1], xy[1:]])

        self.collection = LineCollection(segments, **kwargs)
        self.collection.set_edgecolor(c)


    def set_data(self, x, y, c=None):
        """Set new x and y data, optionally new color data
               x[N]                   xdata
               y[N]                   ydata
               c[N], c[N,3], c[N,4]   color data (default: current colors)
        """

        xy = np.vstack([x,y]).T
        xy = xy.reshape(-1, 1, 2)
        segments = np.hstack([xy[:-1], xy[1:]])

        self.collection.set_segments(segments)
        if c is not None:
            self.collection.set_edgecolor(c)
        
    def set_xdata(self, x):
        """Set new x (must be same size as current y data)
               x[N]                   xdata
        """
        segments = np.array(self.collection.get_segments())
        segments[:,0,0] = x[:-1]
        segments[:,1,0] = x[1:]
        self.collection.set_segments(segments)

    def set_ydata(self, y):
        """Set new y (must be same size as current x data)
               y[N]                   ydata
        """
        segments = np.array(self.collection.get_segments())
        segments[:,0,1] = y[:-1]
        segments[:,1,1] = y[1:]
        self.collection.set_segments(segments)

    def set_colors(self, c):
        """Set new color data
               c[N], c[N,3], c[N,4]   color data
        """
        self.collection.set_edgecolor(c)

    def set_animated(self, animated):
        self.collection.set_animated(animated)

    def get_zorder(self):
        return self.collection.get_zorder()

def colored_plot(x, y, c, ax=None, **kwargs):
    """Like matplotlib plot, but use a list of colors c for variable color. Return a colored_line
            x[N]                   xdata
            y[N]                   ydata
            c[N], c[N,3], c[N,4]   color data
            ax                     axis (default: current axis)
            **kwargs               kwargs valid for matplotlib.LineCollection
    """
    if ax is None:
        ax = plt.gca()
    line = colored_line(x,y,c, **kwargs)

    ax.add_collection(line.collection)
    line.axes = line.collection.axes
    line.figure = line.collection.figure
    line.draw = line.collection.draw

    return line




import numpy as np
from collections.abc import Iterable
import quaternion
from abc import ABCMeta, abstractmethod
from .hydrodynamics import grand_mobility_matrix, particle_wall_self_mobility
from .integrators import basic_integrator
from tqdm import tqdm

class trajectory:
    def __init__(self, position, orientation=None):
        self.position = np.asarray(position, dtype=float)

        if orientation is not None:
            self.orientation = np.asarray(orientation, dtype=np.quaternion)
        else:
            self.orientation = None

    def __getitem__(self, key):
        orientation_s = None if self.orientation is None else self.orientation[key]
        return trajectory(self.position[key], orientation_s)

class interactions:
    """
    Abstract base class for particle interactions
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def force(self):
        """
        Return the force[N,3] based on the current state
        """
        pass

    @abstractmethod
    def torque(self):
        """
        Return the torque[N,3] based on the current state
        """
        pass
    
    @abstractmethod
    def update(self):
        """
        Update the current state
        """
        pass

    def _update(self, time, position, orientation, temperature):
        self.time = time
        self.position = position
        self.orientation = orientation
        self.temperature = temperature
        self.update()

class stokesian_dynamics:
    """
    Perform a Stokesian dynamics simulation, with optional external and internal interactions and rotational dynamics
    """
    def __init__(self, *, temperature, dt, position, drag, orientation=None, 
                 force=None, torque=None, interactions=None, constraint=None,
                 interface=None, inertia=None, hydrodynamic_coupling=True, integrator=None):
        """
        Arguments:
            temperature        system temperature
            dt                 time-step
            position[N,D]      initial position of N particles in D dimensions
            drag               drag coefficients (of base type stoked.drag)
            orientation[N]     initial orientation of N particles (as quaternions)
            force(t, r, q)     external force function given time t, position r[N,D], orientation q[N] and returns force F[N,D] (can be a list of functions)
            torque(t, r, q)    external torque function given time t, position r[N,D], orientation q[N] and returns torque T[N,D] (can be a list of functions)
            interactions       particle interactions (can be a list)
            constraint         particle constraints (can be a list)
            interface          no-slip boundary interface (default: no interface)
            inertia            inertia coefficients of base type stoked.inertia (default: over-damped dynamics)
            hydrodynamic_coupling    if True, include hydrodynamic coupling interactions (default: True)
            integrator         dynamics integrator
        """
        self.temperature = temperature
        self.dt = dt
        self.time = 0
        self.position = np.atleast_2d(np.asarray(position, dtype=float))
        self.drag = drag
        self.Nparticles = len(self.position)
        self.ndim = self.position.shape[1]
        self.interface = interface
        self.inertia = inertia
        self.hydrodynamic_coupling = hydrodynamic_coupling
        self.integrator = integrator

        if integrator is None:
            self.integrator = basic_integrator()

        if self.interface is not None and self.ndim != 3:
            raise ValueError('An interface can only be used in 3-dimensions')

        if not self.drag.isotropic and self.interface is not None:
            raise NotImplementedError('anisotropic particles with an interface')

        if orientation is not None:
            self.orientation = np.atleast_1d(np.asarray(orientation, dtype=np.quaternion))
        else:
            self.orientation = np.ones([self.Nparticles], dtype=np.quaternion)

        if force is None:
            self.force = [lambda t, rvec, orient: np.zeros_like(rvec)]
        elif not isinstance(force, Iterable):
            self.force = [force]
        else:
            self.force = force

        if torque is None:
            self.torque = [lambda t, rvec, orient: np.zeros_like(rvec)]
        elif not isinstance(force, Iterable):
            self.torque = [torque]
        else:
            self.torque = torque

        if interactions is None:
            self.interactions = None
        elif not isinstance(interactions, Iterable):
            self.interactions = [interactions]
        else:
            self.interactions = interactions
        self._combine_pairwise_interactions()

        if constraint is None:
            self.constraint = []
        elif not isinstance(constraint, Iterable):
            self.constraint = [constraint]
        else:
            self.constraint = constraint

        self.alpha_T = np.zeros_like(self.position)
        drag_T = self.drag.drag_T
        if self.drag.isotropic and not np.isscalar(drag_T):
            drag_T = drag_T[:,np.newaxis]

        self.alpha_T[...] = 1/drag_T
        self.velocity = np.zeros_like(self.position)

        if self.orientation is not None:
            self.alpha_R = np.zeros_like(self.position)
            drag_R = self.drag.drag_R
            if self.drag.isotropic and not np.isscalar(drag_R):
                drag_R = drag_R[:,np.newaxis]

            self.alpha_R[...] = 1/drag_R
            self.angular_velocity = np.zeros_like(self.position)

        self.isotropic = self.drag.isotropic and not self.hydrodynamic_coupling
        if self.isotropic and orientation is None:
            self.rotating = False
        else:
            self.rotating = True

        if self.inertia is not None:
            self.mass = np.empty([self.Nparticles], dtype=float)
            self.moment = np.empty([self.Nparticles,3], dtype=float)

            self.mass[...] = self.inertia.mass
            self.moment[...] = self.inertia.moment

        self._init_integrator()

    def step(self):
        """
        Time-step the positions (and orientations) by dt
        """
        self.integrator.pre_step()
        self.step_func()
        self.integrator.post_step()

    def run(self, Nsteps, progress=True):
        """
        Run the simulation for Nsteps, returning the trajectories

        Arguments:
            Nsteps       number of steps to run
            progress     If True, display a progress bar
        """
        position = np.zeros((Nsteps,) + self.position.shape, dtype=float)

        if self.rotating:
            orientation = np.zeros((Nsteps,) + self.orientation.shape, dtype=np.quaternion)
        else:
            orientation = None

        for i in tqdm(range(Nsteps), desc='Running dynamics', disable=(not progress)):
            position[i] = self.position
            if self.rotating:
                orientation[i] = self.orientation
            self.step()

        return trajectory(position, orientation)

    def run_until(self, condition):
        """
        Run the simulation until some condition is met, returning the trajectories

        Arguments:
            condition    Function to return True when simulation should terminate
            progress     If True, display a progress bar
        """
        position = []

        if self.rotating:
            orientation = []
        else:
            orientation = None

        with tqdm(desc='Running dynamics until condition is met', disable=(not progress)) as pbar:
            while not condition():
                position.append(np.copy(self.position))
                if self.rotating:
                    orientation.append(np.copy(self.orientation))
                self.step()
                pbar.update()

        position = np.array(position, dtype=float)
        if self.rotating:
            orientation = np.array(orientation, dtype=np.quaternion)

        return trajectory(position, orientation)

    def _update_interactions(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        update the interactions given a new (or predicted) position and orientation
        """
        if self.interactions is not None:
            for I in self.interactions:
                I._update(time, position, orientation, self.temperature)

    def _total_force(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        return the total force at a given time, position, and orientation
        """
        F = sum((force(time, position, orientation) for force in self.force))
        if self.interactions is not None:
            for I in self.interactions:
                F += I.force()
        return F

    def _total_torque(self, time, position, orientation):
        """
        FOR INTERNAL USE ONLY

        return the total torque at a given time, position, and orientation
        """
        T = sum((torque(time, position, orientation) for torque in self.torque))
        if self.interactions is not None:
            for I in self.interactions:
                T += I.torque()
        return T

    def _combine_pairwise_interactions(self):
        from stoked.forces import pairwise_central_force

        if self.interactions is not None:
            pairwise_forces = []
            for i,I in enumerate(self.interactions):
                if isinstance(I, pairwise_central_force):
                    pairwise_forces.append(I)

            if len(pairwise_forces) > 1:
                for pf in pairwise_forces:
                    self.interactions.remove(pf)
                self.interactions.append(sum(pairwise_forces[1:], pairwise_forces[0]))

    def _init_integrator(self):
        self.integrator.initialize(self)

        if self.hydrodynamic_coupling:
            if self.inertia is None:
                self.step_func = self.integrator.hbd_step
            else:
                self.step_func = self.integrator.hld_step
        else:
            if self.inertia is None:
                self.step_func = self.integrator.bd_step
            else:
                self.step_func = self.integrator.ld_step

def brownian_dynamics(*, temperature, dt, position, drag, orientation=None, 
                 force=None, torque=None, interactions=None, constraint=None, interface=None, inertia=None, integrator=None):
    """
    Perform a Brownian dynamics simulation, with optional external and internal interactions and rotational dynamics
        
    Arguments:
        temperature        system temperature
        dt                 time-step
        position[N,D]      initial position of N particles in D dimensions
        drag               drag coefficients (of base type stoked.drag)
        orientation[N]     initial orientation of N particles (as quaternions)
        force(t, r, q)     external force function given time t, position r[N,D], orientation q[N] and returns force F[N,D] (can be a list of functions)
        torque(t, r, q)    external torque function given time t, position r[N,D], orientation q[N] and returns torque T[N,D] (can be a list of functions)
        interactions       particle interactions (can be a list)
        interface          no-slip boundary interface (default: no interface)
        constraint         particle constraints (can be a list)
        inertia            inertia coefficients of base type stoked.inertia (default: over-damped dynamics)
        integrator         dynamics integrator
    """
    return stokesian_dynamics(temperature=temperature, dt=dt, position=position, drag=drag,
                  orientation=orientation, force=force, torque=torque, interactions=interactions,
                  constraint=constraint, interface=interface, inertia=inertia, hydrodynamic_coupling=False,
                  integrator=integrator)
